/***************************************************************************
 * ggp2mseed.c
 *
 * Timeseries data conversion from GGP to Mini-SEED.
 *
 * Written by Chad Trabant, IRIS Data Management Center
 *
 * modified 2014.093
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <libmseed.h>

#define VERSION "1.1"
#define PACKAGE "ggp2mseed"

/* Scaling applied to floating point values when converting to integer */
#define INTSCALING 1000000

struct metadata {
  int numfields;
  char *fields;
};

struct filelink {
  char *filename;
  struct filelink *next;
};

static int ggp2group (char *ggpfile, MSTraceGroup *mstg);
static char *genblkt2000 (struct metadata *metadata);
static int parameter_proc (int argcount, char **argvec);
static char *getoptval (int argcount, char **argvec, int argopt);
static void addfile (char *filename);
static void record_handler (char *record, int reclen, void *handlerdata);
static char *strclean (char *str);
static void usage (void);

static flag  verbose     = 0;
static int   packreclen  = -1;
static flag  encoding    = 11;
static flag  byteorder   = -1;
static flag  packmeta    = 0;
static char *gravchan    = "UGZ";
static char *preschan    = "UDO";
static char *netcode     = "SG";
static char *loccode     = 0;
static char *outputfile  = 0;
static FILE *ofp         = 0;

struct filelink *filelist = 0;

int
main (int argc, char **argv)
{
  struct filelink *flp;
  MSTraceGroup *mstg = 0;
  MSTrace *mst;
  MSRecord *mstemplate = 0;
  
  int64_t packedsamples = 0;
  int64_t packedrecords = 0;
  
  /* Process given parameters (command line and parameter file) */
  if (parameter_proc (argc, argv) < 0)
    return -1;
  
  /* Init MSTraceGroup */
  mstg = mst_initgroup (mstg);
  
  /* Open the output file if specified, "-" means stdout */
  if ( outputfile )
    {
      if ( strcmp (outputfile, "-") == 0 )
        {
          ofp = stdout;
        }
      else if ( (ofp = fopen (outputfile, "wb")) == NULL )
        {
          fprintf (stderr, "Cannot open output file: %s (%s)\n",
                   outputfile, strerror(errno));
          return -1;
        }
    }
  
  /* Read input GGP files into MSTraceGroup */
  flp = filelist;
  
  while ( flp != 0 )
    {
      if ( verbose )
	fprintf (stderr, "Reading %s\n", flp->filename);

      ggp2group (flp->filename, mstg);
      
      flp = flp->next;
    }

  mst = mstg->traces;
  
  /* Pack data into Mini-SEED records.  If packmeta is true and the private pointer
   * is set add the previously generated Blockette 2000 to a MSRecord template */
  while ( mst )
    {
      int64_t mstsamples;
      int blktlen;
      
      mstemplate = msr_init (mstemplate);
      
      mstemplate->dataquality = 'D';
      strcpy (mstemplate->network, mst->network);
      strcpy (mstemplate->station, mst->station);
      strcpy (mstemplate->location, mst->location);
      strcpy (mstemplate->channel, mst->channel);
      
      if ( packmeta && mst->prvtptr )
	{
	  blktlen = 15 + strlen ((char *) mst->prvtptr + 15);
	  msr_addblockette (mstemplate, (char *) mst->prvtptr, blktlen, 2000, 0);
	}
      
      packedrecords += mst_pack (mst, &record_handler, 0, packreclen, encoding, byteorder,
				 &mstsamples, 1, verbose-1, mstemplate);
      
      packedsamples += mstsamples;
      
      mst = mst->next;
    }
  
  if ( packedrecords < 0 )
    {
      fprintf (stderr, "Error packing data\n");
    }
  else
    {
      fprintf (stderr, "Packed %d trace(s) of %lld samples into %lld records\n",
	       mstg->numtraces, (long long int)packedsamples, (long long int)packedrecords);
    }
  
  /* Make sure everything is cleaned up */
  mst_freegroup (&mstg);
  
  msr_free (&mstemplate);
  
  if ( ofp )
    fclose (ofp);
  
  return 0;
}  /* End of main() */


/***************************************************************************
 * ggp2group:
 *
 * Read a GGP data file and add data samples to a MSTraceGroup.  As the
 * GGP is read in a MSRecord struct is used as a holder for the input
 * information.
 *
 * Returns 0 on success, and -1 on failure
 ***************************************************************************/
static int
ggp2group (char *ggpfile, MSTraceGroup *mstg)
{
  FILE *ifp;
  MSRecord *gmsr = 0;
  MSRecord *pmsr = 0;
  struct metadata metadata;
  char line[1025];
  char *cline;
  int clinesize;
  
  flag stationfound = 0;
  int expectdata = 0;
  int dataend = 0;  /* 1: end of segment, 2: end of block */
  int retval = 0;
  
  hptime_t prevhptime = 0;
  double samprate = 0.0;
  void *gdatabuf = 0;
  void *pdatabuf = 0;
  int32_t databufsize = 0;
  
  /* Checking encoding format sanity */
  if ( encoding != 3 && encoding != 4 && encoding !=5 && encoding != 10 && encoding != 11 )
    {
      fprintf (stderr, "Encoding format %d is not supported\n", encoding);
      return -1;
    }
  
  /* Open input file */
  if ( (ifp = fopen (ggpfile, "rb")) == NULL )
    {
      fprintf (stderr, "Cannot open input file: %s (%s)\n",
	       ggpfile, strerror(errno));
      return -1;
    }
  
  /* Init MSRecord data structures */
  if ( ! (gmsr = msr_init(gmsr)) )
    {
      fprintf (stderr, "Cannot initialize Gravity MSRecord strcture\n");
      return -1;
    }
  if ( ! (pmsr = msr_init(pmsr)) )
    {
      fprintf (stderr, "Cannot initialize Pressure MSRecord strcture\n");
      return -1;
    }
  
  /* Init metadata structure */
  if ( packmeta )
    {
      metadata.numfields = 0;
      metadata.fields = 0;
    }

  while ( fgets (line, 1024, ifp) )
    {
      /* String cleanup, truncate at first newline and cut trailing spaces */
      cline = strclean (line);
      
      if ( cline )
	{
	  clinesize = strlen (cline);
	}
      else
	{
	  fprintf (stderr, "[%s] Error reading input line\n", ggpfile);
	  retval = -1;
	  break;
	}
      
      /* Process data begin lines [C****...] */
      if ( ! strncmp ("C*", cline, 2) )
	{
	  expectdata = 1;
	}
      
      /* Process header, etc. lines */
      else if ( ! expectdata )
	{
	  char *fdesc, *fval;
	  
	  /* Special case */
	  if ( ! strncasecmp ("yyyymmdd", cline, 8) )
	    {
	      continue;
	    }
	  
	  /* Split header line on the colon */
	  if ( ! (fval = strchr (cline, ':')) )
	    {
	      fprintf (stderr, "[%s] No colon found in header: '%s'\n", ggpfile, cline);
	      retval = -1;
	      break;
	    }
	  
	  if ( verbose )
	    fprintf (stderr, "[%s] Header line: %s\n", ggpfile, cline);
	  
	  *(fval++) = '\0';
	  fval = strclean (fval);
	  fdesc = strclean (cline);
	  
	  /* Parse header values */
	  if ( ! strncasecmp ("Filename", fdesc, 8) )
	    {
	      stationfound = 1;
	      
	      if ( netcode )
		{
		  ms_strncpclean (gmsr->network, netcode, 2);
		  ms_strncpclean (pmsr->network, netcode, 2);
		}
	      
	      ms_strncpclean (gmsr->station, fval, 2);
	      ms_strncpclean (pmsr->station, fval, 2);
	      
	      if ( loccode )
		{
		  ms_strncpclean (gmsr->location, loccode, 2);
		  ms_strncpclean (pmsr->location, loccode, 2);
		}
	      
	      ms_strncpclean (gmsr->channel, gravchan, 3);
	      ms_strncpclean (pmsr->channel, preschan, 3);
	    }
	  
	  /* Accumulate metadata (header pair values) */
	  if ( packmeta )
	    {
	      char field[100];
	      
	      /* Create a field string */
	      snprintf (field, 100, "%s:%s~", fdesc, fval);
	      
	      if ( verbose > 2 )
		fprintf (stderr, "Adding '%s' to metadata\n", field);
	      
	      if ( metadata.fields )
		metadata.fields = realloc (metadata.fields, strlen(metadata.fields) + strlen(field) + 1);
	      else
		metadata.fields = realloc (metadata.fields, strlen(field) + 1);
	      
	      strcat (metadata.fields, field);
	      	      
	      metadata.numfields++;
	    }
	}
      
      /* Process step correction lines [77777777] */
      else if ( ! strncmp ("77777777", cline, 8) )
	{
	  if ( verbose > 1 )
	    fprintf (stderr, "[%s] 77777777: Step correction line read\n", ggpfile);
	  
	  if ( ! stationfound )
	    {
	      fprintf (stderr, "[%s] 77777777: Data step correction but station is not yet known!\n",
		       ggpfile);
	      retval = -1;
	      break;
	    }
	}
      
      /* Skip INSTR lines */
      else if ( ! strncmp ("INSTR", cline, 5) )
	{
	  if ( verbose >= 1 )
	    fprintf (stderr, "[%s] INSTR line skipped\n", ggpfile);
	}
      
      /* Process data gap lines [66666666] */
      else if ( ! strncmp ("66666666", cline, 8) )
	{
	  if ( verbose > 1 )
	    fprintf (stderr, "[%s] 66666666: Data gap indicated\n", ggpfile);
	  
	  dataend = 1;
	}
      
      /* Process data end lines [99999999] */
      else if ( ! strncmp ("99999999", cline, 8) )
 	{
	  if ( verbose > 1 )
	    fprintf (stderr, "[%s] 99999999: Data end indicated\n", ggpfile);
	  
	  dataend = 2;
	  expectdata = 0;
	}
      
      /* Process data lines */
      else
	{
	  hptime_t hptime;
	  char datestr[10];
	  char timestr[10];
	  int year, month, mday, yday, hour, min, sec;
	  int ret;
	  
	  char gsample[20];
	  char psample[20];
	  int32_t gint32 = 0;
	  int32_t pint32 = 0;
	  double gdouble = 0.0;
	  double pdouble = 0.0;
	  
	  dataend = 0;
	  
	  if ( verbose > 2 )
	    fprintf (stderr, "Data line: %s\n", cline);
	  
	  if ( clinesize < 35 )
	    {
	      fprintf (stderr, "[%s] %s: Data line is too short, only %d characters:\n%s\n",
		       ggpfile, gmsr->station, clinesize, cline);
	      retval = -1;
	      break;
	    }
	  
	  /* Parse data line, default to a double for integer encoding formats */
	  ret = sscanf (cline, "%s %s %s %s", datestr, timestr, gsample, psample);
	  
	  if ( ret != 4 )
	    {
	      fprintf (stderr, "[%s] %s: Error parsing data line: %s\n",
		       ggpfile, gmsr->station, cline);
	      retval = -1;
	      break;
	    }
	  
	  /* Test for a padding sample and skip if found */
	  if ( ! strcmp (gsample, "99999.999") || ! strcmp (psample, "99999.999") )
	    {
	      if ( verbose )
		fprintf (stderr, "[%s] %s: Skipping gap sample(s): '%s'\n",
			 ggpfile, gmsr->station, cline);
	      
	      dataend = 1;
	    }
	  /* Otherwise add the samples to the accumulation buffers */
	  else
	    {
	      /* Convert the sample values to a double */
	      gdouble = strtod (gsample, NULL);
	      pdouble = strtod (psample, NULL);
	      
	      /* If the encoding format requires integers, scale the doubles */
	      if ( encoding != 4 && encoding != 5 )
		{
		  gint32 = (int32_t) (gdouble * (double)INTSCALING + 0.5);
		  pint32 = (int32_t) (pdouble * (double)INTSCALING + 0.5);
		}
	      
	      /* Convert time stamp */
	      if ( sscanf (datestr, "%4d %2d %2d", &year, &month, &mday) != 3 )
		{
		  fprintf (stderr, "[%s] %s: Error parsing date, skipping: %s\n",
			   ggpfile, gmsr->station, datestr);
		  continue;
		}
	      if ( sscanf (timestr, "%2d %2d %2d", &hour, &min, &sec) != 3 )
		{
		  fprintf (stderr, "[%s] %s: Error parsing time, skipping: %s\n",
			   ggpfile, gmsr->station, timestr);
		  continue;
		}
	      
	      ms_md2doy (year, month, mday, &yday);
	      
	      if ( (hptime = ms_time2hptime (year, yday, hour, min, sec, 0)) == HPTERROR )
		{
		  fprintf (stderr, "[%s] %s: Error converting time stamp to hptime: %s %s\n",
			   ggpfile, gmsr->station, datestr, timestr);
		  retval = -1;
		  break;
		}
	      
	      /* Set start time if not already done */
	      if ( ! gmsr->starttime || ! pmsr->starttime )
		{
		  gmsr->starttime = hptime;
		  pmsr->starttime = hptime;
		}
	      
	      /* Track previous hptime and set and check the sampling rate */
	      if ( prevhptime )
		{
		  samprate = 1.0 / ((double)(hptime - prevhptime)/HPTMODULUS);
		  
		  if ( gmsr->samprate == 0.0 || pmsr->samprate == 0.0 )
		    {
		      gmsr->samprate = samprate;
		      pmsr->samprate = samprate;
		    }
		  else if ( (ms_dabs (1.0 - (gmsr->samprate / samprate))) > 0.0001 )
		    {
		      fprintf (stderr, "[%s] %s: Sampling rate changed from %lf to %lf starting at %s %s\n",
			       ggpfile, gmsr->station,
			       gmsr->samprate, samprate,
			       datestr, timestr);
		      retval = -1;
		      break;
		    }
		}
	      
	      prevhptime = hptime;
	      
	      /* Add data samples to buffer */
	      databufsize++;
	      
	      if ( encoding == 4 ) /* Encoding is 32-bit floats */
		{
		  /* Reallocate sample buffers and add samples */
		  gdatabuf = realloc (gdatabuf, sizeof(float) * databufsize);
		  pdatabuf = realloc (pdatabuf, sizeof(float) * databufsize);
		  
		  *((float *)gdatabuf + (databufsize - 1)) = (float) gdouble;
		  *((float *)pdatabuf + (databufsize - 1)) = (float) pdouble;
		}
	      else if ( encoding == 5 ) /* Encoding is 64-bit doubles */
		{
		  /* Reallocate sample buffers and add samples */
		  gdatabuf = realloc (gdatabuf, sizeof(double) * databufsize);
		  pdatabuf = realloc (pdatabuf, sizeof(double) * databufsize);
		  
		  *((double *)gdatabuf + (databufsize - 1)) = gdouble;
		  *((double *)pdatabuf + (databufsize - 1)) = pdouble;
		}
	      else /* Otherwise encoding requires 32-bit integers */
		{
		  /* Reallocate sample buffers and add samples */
		  gdatabuf = realloc (gdatabuf, sizeof(int32_t) * databufsize);
		  pdatabuf = realloc (pdatabuf, sizeof(int32_t) * databufsize);
		  
		  *((int32_t *)gdatabuf + (databufsize - 1)) = gint32;
		  *((int32_t *)pdatabuf + (databufsize - 1)) = pint32;
		}
	    }
	}
      
      /* If the end of the data is reached process the accumulated data
       * and reset for more data */
      if ( dataend && databufsize > 0 )
	{
	  MSTrace *gtrace, *ptrace;
	  
	  /* Add data to MSTraceGroup */
	  gmsr->datasamples = gdatabuf;
	  gmsr->numsamples = gmsr->samplecnt = databufsize;
	  pmsr->datasamples = pdatabuf;
	  pmsr->numsamples = pmsr->samplecnt = databufsize;
	  
	  if ( encoding == 4 )
	    {
	      gmsr->sampletype = 'f';
	      pmsr->sampletype = 'f';
	    }
	  else if ( encoding == 5 )
	    {
	      gmsr->sampletype = 'd';
	      pmsr->sampletype = 'd';
	    }
	  else
	    {
	      gmsr->sampletype = 'i';
	      pmsr->sampletype = 'i';
	    }
	  
	  if ( verbose >= 1 )
	    {
	      fprintf (stderr, "[%s] %lld samps @ %.6f Hz for N: '%s', S: '%s', L: '%s', C: '%s & %s'\n",
		       ggpfile, (long long int)gmsr->numsamples, gmsr->samprate,
		       gmsr->network, gmsr->station,  gmsr->location,
		       gmsr->channel, pmsr->channel);
	    }
	  
	  if ( ! (gtrace = mst_addmsrtogroup (mstg, gmsr, 0, -1.0, -1.0)) )
	    {
	      fprintf (stderr, "[%s] Error adding gravity samples to MSTraceGroup\n", ggpfile);
	    }
	  
	  if ( ! (ptrace = mst_addmsrtogroup (mstg, pmsr, 0, -1.0, -1.0)) )
	    {
	      fprintf (stderr, "[%s] Error adding pressure samples to MSTraceGroup\n", ggpfile);
	    }
	  
	  /* MSTrace state information in private pointer and cleanup metadata */
	  if ( packmeta )
	    {
	      if ( metadata.fields )
		{
		  gtrace->prvtptr = genblkt2000 (&metadata);
		  ptrace->prvtptr = genblkt2000 (&metadata);
		  
		  /* Free the metadata if this is the end of a block */
		  if ( dataend == 2 )
		    {
		      free (metadata.fields);
		      metadata.fields = 0;
		      metadata.numfields = 0;
		    }
		}
	    }
	  
	  /* Cleanup and reset */
	  if ( dataend == 1 )
	    {
	      gmsr->datasamples = 0;
	      gmsr->starttime = 0;
	      
	      pmsr->datasamples = 0;
	      pmsr->starttime = 0;
	    }
	  else if ( dataend == 2 )
	    {
	      gmsr->datasamples = 0;
	      gmsr = msr_init (gmsr);
	      
	      pmsr->datasamples = 0;
	      pmsr = msr_init (pmsr);
	    }
	  
	  databufsize = 0;
	  prevhptime = 0;
	  dataend = 0;
	}
    } /* Done looping through lines in input file */
  
  if ( packmeta && metadata.fields )
    free (metadata.fields);
  
  if ( gdatabuf )
    free (gdatabuf);
  
  if ( pdatabuf )
    free (pdatabuf);
  
  if ( gmsr )
    msr_free (&gmsr);
  
  if ( pmsr )
    msr_free (&pmsr);
  
  return retval;
}  /* End of gse2group() */


/***************************************************************************
 * genblkt2000:
 *
 * Generate a Blockette 2000 for the given metadata.  The metadata is
 * packed into the blockette as opaque data headers.
 *
 * Returns pointer to blockette body on success and NULL on failure.
 ***************************************************************************/
static char *
genblkt2000 (struct metadata *metadata)
{
  struct blkt_2000_s *blkt2000;
  int headerlen;
  
  if ( ! metadata )
    return NULL;
  
  if ( ! metadata->fields )
    return NULL;
  
  if ( verbose > 1 )
    fprintf (stderr, "Packing %d metadata fields into Blockette 2000\n",
	     metadata->numfields);
  
  headerlen = strlen (metadata->fields);
  
  if ( (blkt2000 = malloc (sizeof(struct blkt_2000_s) + headerlen + 1)) == NULL )
    return NULL;
  
  blkt2000->length = 15 + headerlen + 1;
  blkt2000->data_offset = 15 + headerlen;
  blkt2000->recnum = 1;
  blkt2000->byteorder = 1;
  blkt2000->flags = 0;
  blkt2000->numheaders = metadata->numfields;
  
  /* Copy the opaque header fields, the terminating NULL will be
   * the one byte of opaque data and used later to determine the 
   * blockette body length */
  strncpy (blkt2000->payload, metadata->fields, headerlen + 1);
  
  return (char *) blkt2000;
} /* End of genblkt2000() */


/***************************************************************************
 * parameter_proc:
 * Process the command line parameters.
 *
 * Returns 0 on success, and -1 on failure
 ***************************************************************************/
static int
parameter_proc (int argcount, char **argvec)
{
  int optind;

  /* Process all command line arguments */
  for (optind = 1; optind < argcount; optind++)
    {
      if (strcmp (argvec[optind], "-V") == 0)
	{
	  fprintf (stderr, "%s version: %s\n", PACKAGE, VERSION);
	  exit (0);
	}
      else if (strcmp (argvec[optind], "-h") == 0)
	{
	  usage();
	  exit (0);
	}
      else if (strncmp (argvec[optind], "-v", 2) == 0)
	{
	  verbose += strspn (&argvec[optind][1], "v");
	}
      else if (strcmp (argvec[optind], "-gc") == 0)
	{
	  gravchan = getoptval(argcount, argvec, optind++);
	}
      else if (strcmp (argvec[optind], "-pc") == 0)
	{
	  preschan = getoptval(argcount, argvec, optind++);
	}
      else if (strcmp (argvec[optind], "-n") == 0)
	{
	  netcode = getoptval(argcount, argvec, optind++);
	}
      else if (strcmp (argvec[optind], "-l") == 0)
	{
	  loccode = getoptval(argcount, argvec, optind++);
	}
      else if (strcmp (argvec[optind], "-m") == 0)
	{
	  packmeta = 1;
	}
      else if (strcmp (argvec[optind], "-r") == 0)
	{
	  packreclen = atoi (getoptval(argcount, argvec, optind++));
	}
      else if (strcmp (argvec[optind], "-e") == 0)
	{
	  encoding = atoi (getoptval(argcount, argvec, optind++));
	}
      else if (strcmp (argvec[optind], "-b") == 0)
	{
	  byteorder = atoi (getoptval(argcount, argvec, optind++));
	}
      else if (strcmp (argvec[optind], "-o") == 0)
	{
	  outputfile = getoptval(argcount, argvec, optind++);
	}
      else if (strncmp (argvec[optind], "-", 1) == 0 &&
	       strlen (argvec[optind]) > 1 )
	{
	  fprintf(stderr, "Unknown option: %s\n", argvec[optind]);
	  exit (1);
	}
      else
	{
	  addfile (argvec[optind]);
	}
    }
  
  /* Warning if no output file is specified */
  if ( ! outputfile )
    {
      fprintf (stderr, "Warning! No output file was specified\n");
    }
  
  /* Make sure an input file was specified */
  if ( filelist == 0 )
    {
      fprintf (stderr, "No input files were specified\n\n");
      fprintf (stderr, "%s version %s\n\n", PACKAGE, VERSION);
      fprintf (stderr, "Try %s -h for usage\n", PACKAGE);
      exit (1);
    }

  /* Report the program version */
  if ( verbose )
    fprintf (stderr, "%s version: %s\n", PACKAGE, VERSION);

  return 0;
}  /* End of parameter_proc() */


/***************************************************************************
 * getoptval:
 *
 * Return the value to a command line option; checking that the value is 
 * itself not an option (starting with '-') and is not past the end of
 * the argument list.
 *
 * argcount: total arguments in argvec
 * argvec: argument list
 * argopt: index of option to process, value is expected to be at argopt+1
 *
 * Returns value on success and exits with error message on failure
 ***************************************************************************/
static char *
getoptval (int argcount, char **argvec, int argopt)
{
  if ( argvec == NULL || argvec[argopt] == NULL ) {
    fprintf (stderr, "getoptval(): NULL option requested\n");
    exit (1);
  }
  
  /* Special case of '-o -' usage */
  if ( (argopt+1) < argcount && strcmp (argvec[argopt], "-o") == 0 )
    if ( strcmp (argvec[argopt+1], "-") == 0 )
      return argvec[argopt+1];
  
  if ( (argopt+1) < argcount && *argvec[argopt+1] != '-' )
    return argvec[argopt+1];
  
  fprintf (stderr, "Option %s requires a value\n", argvec[argopt]);
  exit (1);
  return NULL;
}  /* End of getoptval() */


/***************************************************************************
 * addfile:
 *
 * Add file to end of the global file list (filelist).
 ***************************************************************************/
static void
addfile (char *filename)
{
  struct filelink *lastlp, *newlp;
  
  if ( filename == NULL )
    {
      fprintf (stderr, "addfile(): No file name specified\n");
      return;
    }
  
  lastlp = filelist;
  while ( lastlp != 0 )
    {
      if ( lastlp->next == 0 )
        break;
      
      lastlp = lastlp->next;
    }
  
  newlp = (struct filelink *) malloc (sizeof (struct filelink));
  newlp->filename = strdup(filename);
  newlp->next = 0;
  
  if ( lastlp == 0 )
    filelist = newlp;
  else
    lastlp->next = newlp;
  
}  /* End of addfile() */


/***************************************************************************
 * record_handler:
 * Saves passed records to the output file.
 ***************************************************************************/
static void
record_handler (char *record, int reclen, void *handlerdata)
{
  if ( ofp )
    {
      if ( fwrite(record, reclen, 1, ofp) != 1 )
	{
	  fprintf (stderr, "Error writing to output file\n");
	}
    }
}  /* End of record_handler() */


/***************************************************************************
 * strclean:
 * 
 * Truncate string at first newline character, skip to first non-space
 * character and trim spaces from string.
 *
 * Return pointer to string.
 ***************************************************************************/
static char *
strclean (char *str)
{
  char *c;
  
  /* Truncate at first newline character */
  if ( (c = strchr (str, '\n')) )
    {
      *c = '\0';
    }

  /* Skip preceeding spaces */
  while ( *str == ' ' )
    {
      str++;
    }
  
  /* Trim trailing spaces */
  for (c = strrchr(str, '\0') - 1; c >= str; c--)
    {
      if (*c != ' ') break;
      *c = '\0';
    }
  
  return str;
}  /* End of strclean() */


/***************************************************************************
 * usage:
 * Print the usage message and exit.
 ***************************************************************************/
static void
usage (void)
{
  fprintf (stderr, "%s version: %s\n\n", PACKAGE, VERSION);
  fprintf (stderr, "Convert GGP data to Mini-SEED.\n\n");
  fprintf (stderr, "Usage: %s [options] file1 file2 file3...\n\n", PACKAGE);
  fprintf (stderr,
	   " ## Options ##\n"
	   " -V             Report program version\n"
	   " -h             Show this usage message\n"
	   " -v             Be more verbose, multiple flags can be used\n"
	   " -gc channel    Set the SEED gravity channel code, default: %s\n"
	   " -pc channel    Set the SEED pressure channel code, default: %s\n"
	   " -n netcode     Set the SEED network code, default: %s\n"
	   " -l loccode     Set the SEED location code\n"
	   " -m             Pack metadata (headers) into Blockette 2000\n"
	   " -r bytes       Specify record length in bytes for packing, default: 4096\n"
	   " -e encoding    Specify SEED encoding format for packing, default: 11\n"
	   " -b byteorder   Specify byte order for packing, MSBF: 1 (default), LSBF: 0\n"
	   " -o outfile     Specify the output file, use '-' for stdout\n"
	   "\n"
	   " file(s)        File(s) of GGP input data\n"
	   "\n"
	   "Supported Mini-SEED encoding formats:\n"
	   " 3  : 32-bit integers, scaled\n"
	   " 4  : 32-bit floats (C float)\n"
	   " 5  : 64-bit floats (C double)\n"
	   " 10 : Steim 1 compression of scaled 32-bit integers\n"
	   " 11 : Steim 2 compression of scaled 32-bit integers\n"
	   "\n"
	   "For any of the non-floating point encoding formats the data samples\n"
	   "will be scaled by 1,000,000 to avoid precision truncation.\n"
	   "\n", gravchan, preschan, netcode);
}  /* End of usage() */
